#include "connection.hpp"
#include <utility>
#include <boost/log/trivial.hpp>
#include <boost/beast.hpp>
#include "server.hpp"
#include "version.h"

namespace http::server {
    const char* SERVER_HEADER_VALUE = "TerminalWebServer " PROJECT_GIT_REVISION;

    Connection::Connection(boost::asio::ip::tcp::socket socket, ConnectionManager &manager, request_handler handler)
        : socket_(std::move(socket)), connection_manager_(manager), request_handler_(std::move(handler)), request_(false), reply_() {
    }

    void Connection::start() {
        request_parser_.reset();
        request_.reset();
        doRead();
    }

    void Connection::stop() {
        socket_.close();
    }

    Connection::~Connection() = default;

    void Connection::doRead() {
        auto self(shared_from_this());
        socket_.async_read_some(boost::asio::buffer(buffer_), [this, self](boost::system::error_code ec, std::size_t bytes_transferred) {
            if (!ec) {
                RequestParser::result_type result;
                std::tie(result, std::ignore) = request_parser_.parse(
                    request_, buffer_.data(), buffer_.data() + bytes_transferred);

                if (result == RequestParser::good) {
                    request_handler_(request_, reply_);
                    doWrite();
                } else if (result == RequestParser::bad) {
                    stockReply(bad_request, reply_);
                    needClose = true;
                    doWrite();
                } else {
                    doRead();
                }
            } else {
                connection_manager_.stop(shared_from_this());
            }
        });
    }

    void Connection::doWrite() {
        reply_.headers.push_back({.name = "Server", .value = SERVER_HEADER_VALUE});
        reply_.headers.push_back({.name = "Content-Length", .value = std::to_string(reply_.content.size())});
        if (request_.httpVersionMajor == 1) {
            reply_.headers.push_back({.name = "Connection", .value = "keep-alive"});
        }

        const auto ep = socket_.remote_endpoint();
        BOOST_LOG_TRIVIAL(info) << "HTTP query " << ep.address().to_string() << ":" << ep.port() << " " << reply_.status << " " << request_.method << " " << request_.queryUri;

        auto self(shared_from_this());
        async_write(socket_, reply_.to_buffers(), [this, self](boost::system::error_code ec, std::size_t) {
            if (!ec && !needClose) {
                // keep alive Connection
                request_parser_.reset();
                request_.reset();
                doRead();
            } else {
                connection_manager_.stop(shared_from_this());
            }
        });
    }

    SslConnection::SslConnection(boost::asio::ip::tcp::socket socket, ConnectionManager &manager, request_handler handler, const std::shared_ptr<boost::asio::ssl::context>& ctx):
        stream_(std::move(socket), *ctx), connection_manager_(manager), request_handler_(std::move(handler)), request_(true), reply_() {
    }

    void SslConnection::start() {
        get_lowest_layer(stream_).expires_after(std::chrono::seconds(30));
        request_parser_.reset();
        request_.reset();

        // Perform the SSL handshake
        stream_.async_handshake(boost::asio::ssl::stream_base::server, boost::beast::bind_front_handler([this](auto ec) {
            boost::ignore_unused(ec);
            doRead();
        }));
    }

    void SslConnection::stop() {
        try {
            stream_.next_layer().socket().close();
            stream_.shutdown();
        } catch (std::exception& e) {
            BOOST_LOG_TRIVIAL(warning) << "SslConnection::stop(): Can't shutdown ssl socket: " << e.what();
        }
    }

    SslConnection::~SslConnection() = default;

    void SslConnection::doRead() {
        get_lowest_layer(stream_).expires_after(std::chrono::seconds(30));

        auto self(shared_from_this());
        stream_.async_read_some(boost::asio::buffer(buffer_), [this, self](boost::system::error_code ec, std::size_t bytes_transferred) {
            if (!ec) {
                RequestParser::result_type result;
                std::tie(result, std::ignore) = request_parser_.parse(
                    request_, buffer_.data(), buffer_.data() + bytes_transferred);

                if (result == RequestParser::good) {
                    request_handler_(request_, reply_);
                    doWrite();
                } else if (result == RequestParser::bad) {
                    stockReply(bad_request, reply_);
                    needClose = true;
                    doWrite();
                } else {
                    doRead();
                }
            } else {
                connection_manager_.stop(shared_from_this());
            }
        });
    }

    void SslConnection::doWrite() {
        reply_.headers.push_back({.name = "Server", .value = SERVER_HEADER_VALUE});
        reply_.headers.push_back({.name = "Content-Length", .value = std::to_string(reply_.content.size())});
        if (request_.httpVersionMajor == 1) {
            reply_.headers.push_back({.name = "Connection", .value = "keep-alive"});
        }

        const auto ep = stream_.next_layer().socket().remote_endpoint();
        BOOST_LOG_TRIVIAL(info) << "HTTPS query " << ep.address().to_string() << ":" << ep.port() << " " << reply_.status << " " << request_.method << " " << request_.queryUri;

        auto self(shared_from_this());
        async_write(stream_, reply_.to_buffers(), [this, self](boost::system::error_code ec, std::size_t) {
            if (!ec && !needClose) {
                // keep alive Connection
                request_parser_.reset();
                request_.reset();
                doRead();
            } else {
                connection_manager_.stop(shared_from_this());
            }
        });
    }

} // namespace http::Server
